﻿using System;
using UnityEngine;

// This is a VAMDeluxe plugin (see dollmaster)
// Just renamed it to avoid potential namespaces conflicts
// And slightly modified it to avoid triggers with own Autocolliders like the abdomen

namespace VAMMoanPlugin {

    public class TriggerEventArgs:EventArgs
    {
        public Collider collider { get; set; }
        public string evtType { get; set; }
    }

    public class SexTriggerCollide :MonoBehaviour
    {
        TriggerEventArgs lastEvent;

        public event EventHandler<TriggerEventArgs> OnCollide;

        void Awake()
        {
            lastEvent = new TriggerEventArgs
            {
                evtType = "none",
                collider = null
            };
        }

        private void OnTriggerEnter(Collider other)
        {
            DoCollideEvent("Entered", other);
        }

        private void OnTriggerExit(Collider other)
        {
            DoCollideEvent("Exited", other);
        }
        
        /*
        private void OnTriggerStay(Collider other)
        {
            DoCollideEvent("Stay", other);
        }
        */

        private void DoCollideEvent(string evtType,Collider col)
        {
            if (string.Equals(evtType, lastEvent.evtType) && col.gameObject == lastEvent.collider.gameObject)
            {
                return;
            }
            else
            {
				//SuperController.LogMessage( "Collision" + col.gameObject.name );
				// Checking for triggers that are not allowed
				if( col.gameObject.name.Contains("FemaleAutoCollidersabdomen") ) {
					return;
				}
				
                TriggerEventArgs tempEvent = new TriggerEventArgs
                {
                    collider = col,
                    evtType = evtType
                };
                OnCollideEvent(tempEvent);
                lastEvent = tempEvent;
            }
        }

        protected virtual void OnCollideEvent(TriggerEventArgs e)
        {
            EventHandler<TriggerEventArgs> handler = OnCollide;
            handler?.Invoke(this, e);            
        }
        
    }
}