using UnityEngine;
using UnityEngine.UI;
using System;
using System.Collections;
using System.Collections.Generic;
using SimpleJSON;
using System.Linq;
using System.IO;
using MeshVR;
using Request = MeshVR.AssetLoader.AssetBundleFromFileRequest;
using AssetBundles;
using MVR;

// VAMMoan v1.4
//
// Extensively based on Dollmaster plugin and on MacGruber's and AcidBubble's work
// Thank you for all your great plugins guyz !

namespace VAMMoanPlugin
{
    public partial class VAMMoan : MVRScript
    {
        public static string PLUGIN_PATH;
        public static string ASSETS_PATH;
        public static string LOAD_PATH;
					
		private Voice voice;
		private Voices voices;
		private JSONStorableStringChooser voiceChoice;
		
		private JSONStorableString voiceInfos;
		private JSONStorableString stateInfos;
		public JSONStorableStringChooser ArousalMode;
		
		private JSONStorableFloat VAMMVolume;
		private JSONStorableFloat VAMMPitch;
		private JSONStorableFloat VAMMPause;
		private JSONStorableBool VAMMAutoJaw;
		private JSONStorableBool VAMMReverbEnabled;
		private JSONStorableBool VAMMBreathingEnabled;
		private JSONStorableFloat VAMMBreathingScale;
		
		private JSONStorableFloat VAMMKissingSpeed;
		private JSONStorableFloat VAMMBlowjobSpeed;
		
		private JSONStorableFloat maxTriggerCount;
		private JSONStorableFloat arousalRate;
		private JSONStorableFloat TurnOffSpeed;
		private JSONStorableBool bumpUpArousalAfterInactivity;
		private JSONStorableBool allowOrgasm;
		
		private JSONStorableBool enableRandMoanTrigger;
		private JSONStorableFloat randMoanTriggerOccurenceMin;
		private JSONStorableFloat randMoanTriggerOccurenceMax;
		private JSONStorableFloat randMoanTriggerChance;
		private JSONStorableStringChooser randMoanTriggerTypeChoice;
		private JSONStorableString randMoanTriggerHelp;
		
		private string defaultVoice = "Isabella";
		private string selectedVoice;
		private string defaultMode = "Manual";
		private bool voiceInitialized = false;
		
		private float nextMoan = 0f;
		private float nextTurnoffDecrease = 0f;
		private float nextState = 0f;
		private float nextOrgasmCheck = 0f;
		private int moanLoopCount = 0;
		private int moanRandOccurence = 0;

		private float nextTriggerAllowed = 0f; // Used while the orgasm sound is playing, preventing any additionnal increase in arousal or sound played				
		private float lastCollisionTime = -1.0f;
		private float lastInteractionTime = -1.0f;
		private float minLastCollisionTime = 0.8f; // Minimum time in interactive mode to swap currentArousal and play a new intensity instead of breathing

		private float vammTimer = 0f;
				
		private float breathPeriod;
		private float breathStart;
		private float breathEnd;

		private float audioLoudness = 0;
		private float highestLoudness = -10.0f;
		
		private float _currentArousal = -1.0f;
		public float currentArousal
		{
			get
			{
				return _currentArousal;				
			} 
			set
			{
				_currentArousal = value;
				VAMMValue_CurrentArousal.val = value;
			}
		}
		
		private float _sexTriggerCount = 0.0f;
		public float sexTriggerCount
		{
			get
			{
				return _sexTriggerCount;				
			} 
			set
			{
				float newCount = allowOrgasm.val == false && value >= maxTriggerCount.val ? (maxTriggerCount.val - 1) : value;
				_sexTriggerCount = newCount;
				VAMMValue_TriggerCount.val = newCount;
			}
		}
		public float sexTriggerDecayIncrement = 10.0f;
		
		public float defaultArousalIncrement = 0.15f;
		
		private AudioClip currentClipPlaying;
		private AudioClip currentCollisionClipPlaying;
			
		public AudioSourceControl headAudioSource;
		private AdjustJoints JawControlAJ;
		private JSONStorable AutoJawMouthMorphsJS;
		
		string lastAtomName = "";
		
		protected Rigidbody vaTrig;
		
		private bool intensityBoost = false;
				
		// Some buttons and UI stuffs I need globally to swap their states
		private UIDynamicButton showHelpBtn;
		private UIDynamicTextField helpTextfield;
		private UIDynamicButton previewKissingBtn;
		private UIDynamicButton previewBlowjobBtn;
		
		// Morphs ( thank you MacGruber ! )
		DAZCharacterSelector myGeometry;
		DAZCharacterSelector.Gender myGender = DAZCharacterSelector.Gender.None;
		private DAZMorph lifeChestMorph;
		private DAZMorph lifeStomachMorph;
		private DAZMorph lifeNoseOutMorph;
		
		// Actions/triggers
		// ** All starting actions ** ( one shot trigger )
		private EventTrigger startDisabledTrigger;
		private EventTrigger startBreathingTrigger;
		private EventTrigger startKissingTrigger;
		private EventTrigger startBlowjobTrigger;
		private EventTrigger startIntensity0Trigger;
		private EventTrigger startIntensity1Trigger;
		private EventTrigger startIntensity2Trigger;
		private EventTrigger startIntensity3Trigger;
		private EventTrigger startIntensity4Trigger;
		
		private EventTrigger intensityLoweredTrigger;
		private EventTrigger intensityIncreasedTrigger;
		private EventTrigger reachOrgasmTrigger;
		private EventTrigger endOrgasmTrigger;
		
		private List<EventTrigger> allTriggers;
		
		private string editTriggerListDefault = "Select a trigger to edit";
		private List<string> editTriggersList;
		private JSONStorableStringChooser editTriggerChoice;
		
		// ** All transition actions ( updated every frame )
		private EventTrigger breathingTrigger;
		
		// All accessible Storable for scripting purposes
		public JSONStorableFloat VAMMValue_CurrentArousal;
		public JSONStorableFloat VAMMValue_TriggerCount;
		public JSONStorableFloat VAMMValue_MaxTriggerCount;
		
		public override void Init()
        {			 
            try
            {
                if (containingAtom.type != "Person")
                {
                    SuperController.LogError("Please add VAMMoan to a Person atom");
                    return;
                }

				vaTrig = containingAtom.rigidbodies.First(rb => rb.name == "VaginaTrigger");
				vaTrig.gameObject.AddComponent<SexTriggerCollide>().OnCollide += VaTrigObserver;
							
                lastAtomName = containingAtom.uid;
                
				PLUGIN_PATH = GetPluginPath(this);
				ASSETS_PATH = PLUGIN_PATH + "/audio";
				               
                LOAD_PATH = SuperController.singleton.currentLoadDir;
				
				JawControlAJ = containingAtom.GetStorableByID("JawControl") as AdjustJoints;
				AutoJawMouthMorphsJS = containingAtom.GetStorableByID("AutoJawMouthMorph") as JSONStorable;
				
				headAudioSource = containingAtom.GetStorableByID("HeadAudioSource") as AudioSourceControl;
				headAudioSource.SetFloatParamValue("delayBetweenQueuedClips", 0.0f);
				
				initAttenuation();
				
				selectedVoice = defaultVoice;
				voices = new Voices();
				
				/* ********************************* */
				/* **** LEFT UI                 **** */
				/* ********************************* */
				voiceChoice = new JSONStorableStringChooser("voice", voices.voicesNames, defaultVoice, "Voice", (string choice)=>
				{
					selectedVoice = choice;
					ChangeVoice();
				});

				RegisterStringChooser(voiceChoice);
				UIDynamicPopup voicePopup = CreateScrollablePopup(voiceChoice);
				voicePopup.popup.topButton.image.color = new Color(0.35f, 0.60f, 0.65f);
				voicePopup.popup.selectColor = new Color(0.35f, 0.60f, 0.65f);

				voiceInfos = new JSONStorableString("VoiceInfos", "");
				UIDynamic voiceinfosTextfield = CreateTextField(voiceInfos, false);
				voiceinfosTextfield.height = 200.0f;
				
				stateInfos = new JSONStorableString("Stateinfos", "");
				UIDynamic stateinfosTextfield = CreateTextField(stateInfos, false);
				stateinfosTextfield.height = 90.0f;

				List<string> ModesList = new List<string>();
				ModesList.Add("Manual");
				ModesList.Add("Interactive");
				
				ArousalMode = new JSONStorableStringChooser("arousalMode", ModesList, defaultMode, "Mode", switchArousalModeCallback);
				RegisterStringChooser(ArousalMode);
				UIDynamicPopup arousalModePopup = CreatePopup(ArousalMode);
				arousalModePopup.popup.topButton.image.color = new Color(0.35f, 0.60f, 0.65f);
				arousalModePopup.popup.selectColor = new Color(0.35f, 0.60f, 0.65f);
				
				JSONStorableString helpText = new JSONStorableString("Help",
					"<color=#000><size=35><b>VAMMoan help</b></size></color>\n\n" + 
					"<color=#333>" +
					"<b>Voice:</b> The currently selected girl voice.\n" +
					"<b>State:</b> Shows the current state of arousal selected.\n" +
					"<b>Mode:</b> The current mode of voice playback.\n" +
					"<size=25>- Manual : you have to drive the intensity with actions and triggers\n" +
					"- Interactive : playing with the girl will automatically arouse her (does not work for male)</size>\n\n" +
					"<color=#000><size=35><b>Global Options</b></size></color>\n" + 
					"<i><size=25>Options for both manual and interactive mode</size></i>\n\n" +
					"<b>Voice volume :</b> changes the head audio source volume.\n<i>Warning:</i> reducing the volume will lower the jaw motion if it is enabled.\n" +
					"<b>Voice pitch :</b> changes the head audio pitch (makes a lower or higher voice).\n\n" +
					"<b>Enable Reverb (VAMAtmosphere) :</b> enables the proper configuration of the head audio to allow reverb using VAMAtmosphere. Check this if you want the voice to be affected by the Reverb Zones. <b>You need VAMAtmosphere plugin to add Reverb Zone</b>.\n\n" +
					"<b>Enable breathing animation :</b> enables morph animations to simulate breathing. <b>You need to have MacGruber's Life installed. Compatible versions are 10 to 13</b>.\n\n" +
					"<b>Enable auto-jaw animation :</b> enables the automatic animated jaw based on the voice audio. Check it to enable it, and click once on <b>Set optimal settings</b> to use predefined settings for this voice. You can tweak them aftewards if you want in the <b>Jaw Physics & triggers</b> tab.\n\n" +
					"<color=#000><size=35><b>Manual Options</b></size></color>\n" + 
					"<i><size=25>Options for manual mode only</size></i>\n\n" +
					"<b>Kissing speed :</b> it changes the delay between the kisses / smooches sounds. The lower the value, the faster they get.\n" +
					"<b>Blowjob speed :</b> it changes the delay between the sucking sounds. The lower the value, the faster they get.\n\n" +
					"<color=#000><size=35><b>Interactive Options</b></size></color>\n" + 
					"<i><size=25>Options for interactive mode only</size></i>\n\n" +
					"<b>Raise arousal when resuming: </b> If you stop interacting with the girl for a few seconds, and you resume playing with her. If this checkbox is enabled, there is 25% chance that she will be super excited and get to the next level of intensity immediately.\n" +
					"<b>Allow orgasm: </b> If checked, the girl can have an orgasm. If not, she will stay at intensity 4 if you keep playing with her.\n" +
					"<b>Hard to please: </b> If the girl is hard to please or not. A low value is a lewd girl reaching orgasm pretty fast. A higher value is a frigid girl that demands of bit of trust to please.\n" +
					"<b>Arousal rate: </b> The amount of arousal added during interaction. It can be used to simulate situations where the girl is more or less sensitive when you interact with her.\n" +
					"<b>Turn Off Speed: </b> How much the girl is turned of when you don't interact with her. If the value is low, she will stay excited longer. If the value is high, she will be turned off pretty quick.\n" +
					"</color>"
				);
				helpTextfield = CreateTextField(helpText, false);
				helpTextfield.height = 650.0f;
				
				CreateSpacer();

				// TITLE TESTS
				createStaticDescriptionText("Tests states","<color=#000><size=35><b>TEST STATES</b></size></color>\n<size=24>Use these buttons below to test the different state of the character. Only works in manual mode.</size>",false,130);
				
				
				UIDynamicButton previewDisabledBtn = CreateButton("Disabled", false);
				if (previewDisabledBtn != null) {
					previewDisabledBtn.button.onClick.AddListener( () => { setArousal( -10.0f ); } );
				}

				UIDynamicButton previewBreathingBtn = CreateButton("Breathing", false);
				if (previewBreathingBtn != null) {
					previewBreathingBtn.button.onClick.AddListener( () => { setArousal( -1.0f ); } );
				}
				
				previewKissingBtn = CreateButton("Kissing", false);
				if (previewKissingBtn != null) {
					previewKissingBtn.button.onClick.AddListener( () => { setArousal( 100.0f ); } );
				}
				
				previewBlowjobBtn = CreateButton("Blowjob", false);
				if (previewBlowjobBtn != null) {
					previewBlowjobBtn.button.onClick.AddListener( () => { setArousal( 101.0f ); } );
				}
				
				UIDynamicButton previewNeutralBtn = CreateButton("Intensity 0", false);
				if (previewNeutralBtn != null) {
					previewNeutralBtn.button.onClick.AddListener( () => { setArousal( 0.0f ); } );
				}
				
				UIDynamicButton previewIntensity1Btn = CreateButton("Intensity 1", false);
				if (previewIntensity1Btn != null) {
					previewIntensity1Btn.button.onClick.AddListener( () => { setArousal( 1.0f ); } );
				}
				
				UIDynamicButton previewIntensity2Btn = CreateButton("Intensity 2", false);
				if (previewIntensity2Btn != null) {
					previewIntensity2Btn.button.onClick.AddListener( () => { setArousal( 2.0f ); } );
				}
				
				UIDynamicButton previewIntensity3Btn = CreateButton("Intensity 3", false);
				if (previewIntensity3Btn != null) {
					previewIntensity3Btn.button.onClick.AddListener( () => { setArousal( 3.0f ); } );
				}
				
				UIDynamicButton previewIntensity4Btn = CreateButton("Intensity 4", false);
				if (previewIntensity4Btn != null) {
					previewIntensity4Btn.button.onClick.AddListener( () => { setArousal( 4.0f ); } );
				}
				
				UIDynamicButton previewIntensity5Btn = CreateButton("Orgasm", false);
				if (previewIntensity5Btn != null) {
					previewIntensity5Btn.button.onClick.AddListener( () => { setArousal( 5.0f ); } );
				}
				
				/* ********************************* */
				/* **** RIGHT UI                **** */
				/* ********************************* */

				// TITLE GLOBAL
				createStaticDescriptionText("Global options","<color=#000><size=35><b>GLOBAL OPTIONS</b></size></color>",true,40);
				
				
				VAMMVolume = new JSONStorableFloat("Voice volume", 1f, VAMMVolumeCallback, 0f, 1f);
				RegisterFloat(VAMMVolume);
				CreateSlider(VAMMVolume, true);
				VAMMVolume.val = headAudioSource.GetFloatParamValue("volume");
				
				VAMMPitch = new JSONStorableFloat("Voice pitch", 1f, VAMMPitchCallback, 0.1f, 3f);
				RegisterFloat(VAMMPitch);
				CreateSlider(VAMMPitch, true);
				VAMMPitch.val = headAudioSource.GetFloatParamValue("pitch");
				
				VAMMReverbEnabled = new JSONStorableBool("Enable Reverb (VAMAtmosphere)", false, VAMMReverbEnabledCallback);
				VAMMReverbEnabled.storeType = JSONStorableParam.StoreType.Full;
				CreateToggle(VAMMReverbEnabled, true);
				RegisterBool(VAMMReverbEnabled);
				VAMMReverbEnabled.val = !headAudioSource.GetBoolParamValue("spatialize");
				
				VAMMBreathingEnabled = new JSONStorableBool("Enable breathing animation", true, VAMMBreathingEnabledCallback);
				VAMMBreathingEnabled.storeType = JSONStorableParam.StoreType.Full;
				CreateToggle(VAMMBreathingEnabled, true);
				RegisterBool(VAMMBreathingEnabled);
				
				VAMMAutoJaw = new JSONStorableBool("Enable auto-jaw animation", false, VAMMAutoJawCallback);
				VAMMAutoJaw.storeType = JSONStorableParam.StoreType.Full;
				CreateToggle(VAMMAutoJaw, true);
				RegisterBool(VAMMAutoJaw);
				VAMMAutoJaw.val = JawControlAJ.GetBoolParamValue("driveXRotationFromAudioSource"); // Reading the parameters from the normal tabs
				AutoJawMouthMorphsJS.SetBoolParamValue("enabled", VAMMAutoJaw.val); // Also settings the morphs to the same value
				
				UIDynamicButton setOptimalJawParams = CreateButton("Set optimal auto-jaw animation parameters", true);
				if (setOptimalJawParams != null) {
					setOptimalJawParams.button.onClick.AddListener( setOptimalJawParamsCallback );
				}
				
				CreateSpacer(true);
				
				// TITLE MANUAL
				createStaticDescriptionText("Manual mode options","<color=#000><size=35><b>MANUAL MODE</b></size></color>",true,40);

				
				VAMMKissingSpeed = new JSONStorableFloat("Kissing Speed", 0.50f, 0f, 1f);
				RegisterFloat(VAMMKissingSpeed);
				CreateSlider(VAMMKissingSpeed, true);
				
				VAMMBlowjobSpeed = new JSONStorableFloat("Blowjob Speed", 1f, 0.5f, 1f);
				RegisterFloat(VAMMBlowjobSpeed);
				CreateSlider(VAMMBlowjobSpeed, true);
		
				CreateSpacer(true);
				
				// TITLE AUTO
				createStaticDescriptionText("Auto mode options","<color=#000><size=35><b>INTERACTIVE MODE</b></size></color>",true,40);

				
				bumpUpArousalAfterInactivity = new JSONStorableBool("Raise arousal when resuming", true);
				bumpUpArousalAfterInactivity.storeType = JSONStorableParam.StoreType.Full;
				CreateToggle(bumpUpArousalAfterInactivity, true);
				RegisterBool(bumpUpArousalAfterInactivity);
				
				allowOrgasm = new JSONStorableBool("Allow orgasm", true);
				allowOrgasm.storeType = JSONStorableParam.StoreType.Full;
				CreateToggle(allowOrgasm, true);
				RegisterBool(allowOrgasm);
				
				// Default at 700 is arbitrary. It feels like a good middle ground for a quick interactive experience
				maxTriggerCount = new JSONStorableFloat("Hard to please", 700.0f, (val) => { maxTriggerCount.valNoCallback = Mathf.Round(val); maxTriggerCountCallback( maxTriggerCount.val ); }, 500.0f, 3000.0f, true, true);
				maxTriggerCount.storeType = JSONStorableParam.StoreType.Full;
				CreateSlider(maxTriggerCount, true);
				RegisterFloat(maxTriggerCount);
				
				arousalRate = new JSONStorableFloat("Arousal rate", 1.0f, arousalRateCallback, 0.1f, 3.0f, true, true);
				CreateSlider(arousalRate, true);
				RegisterFloat(arousalRate);
				
				TurnOffSpeed = new JSONStorableFloat("Turn Off Speed", 1.5f, 1f, 2.5f, true, true);
				TurnOffSpeed.storeType = JSONStorableParam.StoreType.Full;
				CreateSlider(TurnOffSpeed, true);
				RegisterFloat(TurnOffSpeed);
				
				
				CreateSpacer(true);
				
				// Creating my triggers
				initTriggers();
				
				// TITLE ADVANCED
				createStaticDescriptionText("Advanced options","<color=#000><size=35><b>ADVANCED OPTIONS</b></size></color>",true,40);

				// Breathing scale
				VAMMBreathingScale = new JSONStorableFloat("Breathing scale", 1f, VAMMBreathingScaleCallback, 0.1f, 1f);
				RegisterFloat(VAMMBreathingScale);
				CreateSlider(VAMMBreathingScale, true);

				createStaticDescriptionText("Breathing scale description","<color=#000><size=26><i><b>Changes how much the breathing animations scales.</b> 1 means it will be animated to it's maximum depth.</i></size></color>",true,95);
				
				// Pause between samples
				VAMMPause = new JSONStorableFloat("Pause between moans", 0f, VAMMPauseCallback, 0f, 1.5f);
				CreateSlider(VAMMPause, true);
				
				createStaticDescriptionText("Pause description","<color=#000><size=26><i><b>Adds a tiny additional pause between moans.</b> 0 means no pause, 1 is a one second pause. It can be used to slow down the pace of the moans during an animation.</i></size></color>",true,125);
				
				CreateSpacer(true);
				
				// Random Moans trigger
				enableRandMoanTrigger = new JSONStorableBool("Enable randomized moan playback", false);
				enableRandMoanTrigger.storeType = JSONStorableParam.StoreType.Full;
				CreateToggle(enableRandMoanTrigger, true);
				RegisterBool(enableRandMoanTrigger);
				
				randMoanTriggerOccurenceMin = new JSONStorableFloat("Occurence Min", 3.0f, (val) => { randMoanTriggerOccurenceMin.valNoCallback = Mathf.Round(val); randMoanTriggerOccurenceMinCallback(val); randMoanTriggerOccurenceMax.SetVal( Mathf.Max(randMoanTriggerOccurenceMax.val, val) ); }, 1.0f, 20.0f);
				RegisterFloat(randMoanTriggerOccurenceMin);
				CreateSlider(randMoanTriggerOccurenceMin, true);
				
				randMoanTriggerOccurenceMax = new JSONStorableFloat("Occurence Max", 8.0f, (val) => { randMoanTriggerOccurenceMax.valNoCallback = Mathf.Round(val); randMoanTriggerOccurenceMaxCallback(val); randMoanTriggerOccurenceMin.SetVal( Mathf.Min(randMoanTriggerOccurenceMin.val, val) ); }, 1.0f, 20.0f);
				RegisterFloat(randMoanTriggerOccurenceMax);
				CreateSlider(randMoanTriggerOccurenceMax, true);
				
				randMoanTriggerChance = new JSONStorableFloat("Playback chance", 0.85f, randMoanTriggerChanceCallback, 0.1f, 1f);
				RegisterFloat(randMoanTriggerChance);
				CreateSlider(randMoanTriggerChance, true);

				List<string> rndMoanTriggerList = new List<string>();
				rndMoanTriggerList.Add("Breathing");
				rndMoanTriggerList.Add("Intensity 0");
				rndMoanTriggerList.Add("Intensity 1");
				rndMoanTriggerList.Add("Intensity 2");
				rndMoanTriggerList.Add("Intensity 3");
				rndMoanTriggerList.Add("Intensity 4");
				rndMoanTriggerList.Add("Current Intensity -1");
				rndMoanTriggerList.Add("Current Intensity -2");
				rndMoanTriggerList.Add("Current Intensity +1");
				rndMoanTriggerList.Add("Current Intensity +2");
				rndMoanTriggerList.Add("Current Intensity +/-1");
				rndMoanTriggerList.Add("Current Intensity +/-2");
				rndMoanTriggerList.Add("Current Intensity +/-X");
				randMoanTriggerTypeChoice = new JSONStorableStringChooser("Playback Type", rndMoanTriggerList, "Current Intensity -1", "Playback Type");
				RegisterStringChooser(randMoanTriggerTypeChoice);
				UIDynamicPopup randMoanTriggerPopup = CreateScrollablePopup(randMoanTriggerTypeChoice,true);
				randMoanTriggerPopup.popup.topButton.image.color = new Color(0.35f, 0.60f, 0.65f);
				randMoanTriggerPopup.popup.selectColor = new Color(0.35f, 0.60f, 0.65f);
				
				randMoanTriggerHelp = createStaticDescriptionText("Rand trigger description","",true,195);
				
				CreateSpacer(true);
				
				editTriggersList = new List<string>();
				editTriggersList.Add("Select a trigger to edit");
				editTriggersList.Add("Start disabled");
				editTriggersList.Add("Start breathing");
				editTriggersList.Add("Start kissing");
				editTriggersList.Add("Start blowjob");
				editTriggersList.Add("Start intensity 0");
				editTriggersList.Add("Start intensity 1");
				editTriggersList.Add("Start intensity 2");
				editTriggersList.Add("Start intensity 3");
				editTriggersList.Add("Start intensity 4");
				editTriggersList.Add("When intensity lowered");
				editTriggersList.Add("When intensity increased");
				editTriggersList.Add("Start orgasm");
				editTriggersList.Add("End orgasm");
				editTriggersList.Add("While breathing");
				
				editTriggerChoice = new JSONStorableStringChooser("triggers", editTriggersList, editTriggerListDefault, "Triggers", (string choice)=>
				{
					switch( choice ) {
						case "Start disabled":
							startDisabledTrigger.OpenPanelActionStart();
							break;
						case "Start breathing":
							startBreathingTrigger.OpenPanelActionStart();
							break;
						case "Start kissing":
							startKissingTrigger.OpenPanelActionStart();
							break;
						case "Start blowjob":
							startBlowjobTrigger.OpenPanelActionStart();
							break;
						case "Start intensity 0":
							startIntensity0Trigger.OpenPanelActionStart();
							break;
						case "Start intensity 1":
							startIntensity1Trigger.OpenPanelActionStart();
							break;
						case "Start intensity 2":
							startIntensity2Trigger.OpenPanelActionStart();
							break;
						case "Start intensity 3":
							startIntensity3Trigger.OpenPanelActionStart();
							break;
						case "Start intensity 4":
							startIntensity4Trigger.OpenPanelActionStart();
							break;
						case "When intensity lowered":
							intensityLoweredTrigger.OpenPanelActionStart();
							break;
						case "When intensity increased":
							intensityIncreasedTrigger.OpenPanelActionStart();
							break;
						case "Start orgasm":
							reachOrgasmTrigger.OpenPanelActionStart();
							break;
						case "End orgasm":
							endOrgasmTrigger.OpenPanelActionStart();
							break;
						case "While breathing":
							breathingTrigger.OpenPanelTransition();
							break;
					}
					editTriggerChoice.val = editTriggerListDefault;
				}){ isStorable = false };
				RegisterStringChooser(editTriggerChoice);
				UIDynamicPopup editTriggerPopup = CreateScrollablePopup(editTriggerChoice,true);
				editTriggerPopup.popup.topButton.image.color = new Color(0.35f, 0.60f, 0.65f);
				editTriggerPopup.popup.selectColor = new Color(0.35f, 0.60f, 0.65f);
				
				// Fake spacers because meshed does not take in account the size of the popup :/
				CreateSpacer(true); CreateSpacer(true); CreateSpacer(true); CreateSpacer(true); CreateSpacer(true); CreateSpacer(true); CreateSpacer(true); CreateSpacer(true);

				// ***************************************************************
				// ******* ACTIONS ***********************************************
				// ***************************************************************
				JSONStorableAction setVoiceDisabled = new JSONStorableAction("Voice disabled", () =>
				{
					setArousal( -10.0f );
				});
				
				JSONStorableAction setVoiceBreathing = new JSONStorableAction("Voice breathing", () =>
				{
					setArousal( -1.0f );
				});

				JSONStorableAction setVoiceKissing = new JSONStorableAction("Voice kissing", () =>
				{
					setArousal( 100.0f );
				});
				
				JSONStorableAction setVoiceBlowjob = new JSONStorableAction("Voice blowjob", () =>
				{
					setArousal( 101.0f );
				});
				
				JSONStorableAction setVoiceI0 = new JSONStorableAction("Voice intensity 0", () =>
				{
					setArousal( 0.0f );
				});
				
				JSONStorableAction setVoiceI1 = new JSONStorableAction("Voice intensity 1", () =>
				{
					setArousal( 1.0f );
				});
				
				JSONStorableAction setVoiceI2 = new JSONStorableAction("Voice intensity 2", () =>
				{
					setArousal( 2.0f );
				});
				
				JSONStorableAction setVoiceI3 = new JSONStorableAction("Voice intensity 3", () =>
				{
					setArousal( 3.0f );
				});
				
				JSONStorableAction setVoiceI4 = new JSONStorableAction("Voice intensity 4", () =>
				{
					setArousal( 4.0f );
				});
				
				JSONStorableAction setVoiceOrg = new JSONStorableAction("Voice orgasm", () =>
				{
					setArousal( 5.0f );
				});
				
				JSONStorableAction enableIntensityBoost = new JSONStorableAction("Enable Intensity Boost", () =>
				{
					intensityBoost = true;
				});
				
				JSONStorableAction disableIntensityBoost = new JSONStorableAction("Disable Intensity Boost", () =>
				{
					intensityBoost = false;
				});
				
				
				JSONStorableAction fakeFuncUseBelow = new JSONStorableAction("- - - - Use these functions below ↓ - - - - -", () => {});
				
				
				// **********************************************************************************************************************************
				// Registering variables, a bit strange to disconnect them from the initial creation, but allows me to order them in the action list
				// **********************************************************************************************************************************
				RegisterAction(fakeFuncUseBelow);
				RegisterAction(setVoiceDisabled);
				RegisterAction(setVoiceBreathing);
				RegisterAction(setVoiceKissing);
				RegisterAction(setVoiceBlowjob);
				RegisterAction(setVoiceI0);
				RegisterAction(setVoiceI1);
				RegisterAction(setVoiceI2);
				RegisterAction(setVoiceI3);
				RegisterAction(setVoiceI4);
				RegisterAction(setVoiceOrg);
				RegisterAction(enableIntensityBoost);
				RegisterAction(disableIntensityBoost);
				
				RegisterFloat(VAMMPause);
				
				// ***************************************************************
				// ******* VALUES ************************************************
				// ***************************************************************
				
				JSONStorableAction fakeFuncUseBelow2 = new JSONStorableAction("- - - - Vars below are meant to be read by external plugins ↓ - - - - -", () => {});
				VAMMValue_CurrentArousal = new JSONStorableFloat("VAMM CurrentArousal", -1f, -1f, 4f){ isStorable = false };
				VAMMValue_TriggerCount = new JSONStorableFloat("VAMM TriggerCount", 0f, 0f, 5000f){ isStorable = false };
				VAMMValue_MaxTriggerCount = new JSONStorableFloat("VAMM MaxTriggerCount", 700f, 500f, 1500f){ isStorable = false };
				
				RegisterAction(fakeFuncUseBelow2);
				RegisterFloat(VAMMValue_CurrentArousal);
				RegisterFloat(VAMMValue_TriggerCount);
				RegisterFloat(VAMMValue_MaxTriggerCount);

				
				// Initializing my custom animation curves
				initAnimationCurves();
				
				// Other inits
				moanRandOccurence = (int)UnityEngine.Random.Range(randMoanTriggerOccurenceMin.val, randMoanTriggerOccurenceMax.val);
				updateRandomizedMoanHelp();
            }
            catch(Exception e)
            {
                SuperController.LogError("VAMMOAN - Exception caught: " + e);
            }
		}
		
		void Start()
        {				
			setArousal(currentArousal, true);
        }
		
		protected void FixedUpdate(){		
			// **********************************
			// Initialization stuffs start
			// **********************************
			
			// We don't apply the moan logic while the scene is loading OR when we have freeze enabled
			if ( SuperController.singleton.isLoading || SuperController.singleton.freezeAnimation ) {
				return;
			}
			
			// If the voice isn't initialized, we do that when the scene has finished loading
			if ( !SuperController.singleton.isLoading && voiceInitialized == false ) {
				ChangeVoice();
										
				myGeometry = containingAtom.GetStorableByID("geometry") as DAZCharacterSelector;
				myGender = myGeometry != null ? myGeometry.gender : DAZCharacterSelector.Gender.None;
				GenerateDAZMorphsControlUI morphControl = myGeometry?.morphsControlUI;
				if (morphControl != null) {
					// Looking for Life from version 13 to 10
					int lifeVersion = 13; 
					while( lifeChestMorph == null ) {
						
						string morphDirFemale = "MacGruber.Life." + lifeVersion + ":/Custom/Atom/Person/Morphs/female/MacGruber/Pose/Breathing/";
						string morphDirMale = "MacGruber.Life." + lifeVersion + ":/Custom/Atom/Person/Morphs/male/MacGruber/Pose/Breathing/";

						string morphDir = myGender == DAZCharacterSelector.Gender.Male ? morphDirMale : morphDirFemale;

						lifeChestMorph = morphControl.GetMorphByUid(morphDir + "Breathing_Chest.vmi");
						lifeStomachMorph = morphControl.GetMorphByUid(morphDir + "Breathing_Stomach.vmi");
						lifeNoseOutMorph = morphControl.GetMorphByUid(morphDir + "Breathing_NoseOut.vmi");

						lifeVersion--;
						if( lifeVersion < 10 ) break;
					}
				}

			}
				
			// **********************************
			// Initialization stuffs end
			// **********************************
			
			// Internal timer
			vammTimer += Time.fixedDeltaTime;
			
			// Updating all the morphs
			updateMorphs();
			
			// Interactive arousal
			float intAr = getInteractiveArousal();
			
			// Controlling state infos and override that are not really important to update every frame
			if( vammTimer >= nextState )
            {
				// Pitch, if it has changed in the headAudioSource, I'm overriding it here
				if( headAudioSource.GetFloatParamValue("pitch") != VAMMPitch.val ) {
					VAMMPitchCallback( VAMMPitch.val );
				}
				
				// State infos
				nextState = vammTimer + 0.7f;
				stateInfos.val = getStateInfos();
			}
			
			 // Voice disabled, we don't do nothin'
			if( currentArousal == -10.0f ) return;
			
			// Before anything else, I'm checking the orgasm state every 200ms
			// because i'm gonna reset a few variables and delay all the sound that are going to play
			if( vammTimer >= nextOrgasmCheck && currentArousal != 10000.0f && ( ArousalMode.val == "Manual" || ( ArousalMode.val == "Interactive" && allowOrgasm.val == true ) ) ) {
				nextOrgasmCheck = vammTimer + 0.2f;
				if( intAr >= 5.0f || currentArousal == 5.0f ) {
					AudioClip oClip = currentClipPlaying = voice.GetTriggeredAudioOrgasm();
					
					if( oClip != null ) {
						// Executing triggers configured for the orgasm
						executeTriggers(5f);
						
						headAudioSource.CallAction("ClearQueue");
						headAudioSource.Stop();
						nextMoan = nextTriggerAllowed = vammTimer + ( oClip.length / Math.Abs(VAMMPitch.val) );
						sexTriggerCount = 0;
						currentArousal = 10000.0f; // Value that will be used to reset at the next cycle
						
						PlayAudio(oClip);
						
						breathStart = vammTimer;
						breathEnd = nextMoan;
					}
				}
			}
	
			// Controlling the sound : moans and shared (kisses, blowjob)
			if( vammTimer >= nextMoan )
            {
				// Reseting to 0 if we were waiting for the orgasm to finish
				if( currentArousal == 10000.0f ) {
					currentArousal = 0f; // Manually, and not through setArousal function 'coz we need to override the behavior of the function
					endOrgasmTrigger.Trigger(); // Manually, this is a specific case, the orgasm just ended
					executeTriggers(0f); // The the other cases
				}
				
				// I'm gonna drive the interactive mode with the "manual" mode
				// The reason behind that is that, if the sounds are triggered immediately from a collision
				// it is really hard to find a good pace or a nice solution to have a pretty seamless and credible voice
				// So since the samples are pretty short, it only delays the audio reaction based on the
				// interaction by 0.2sec to 1sec... tested in VR, it almost un-noticable especially if you don't focus on that
				if( ArousalMode.val == "Interactive" ) {
					float normalizedArousal = (float)Math.Floor( getInteractiveArousal() );
					float lastCollisionDiff = vammTimer - lastCollisionTime;
					float lastInteractionDiff = vammTimer - lastInteractionTime;
									
					// We either get the intensity we want based on the amount of interaction we did
					if( lastCollisionDiff <= minLastCollisionTime ) {
						// Let's imagine we've started to play with the girl, and we stopped a few moments...
						// We're gonna bump up the next moan to the next intensity one time like if she was more excited because you did not touch her enough
						if( lastInteractionTime != -1.0f ) { // First of all, only if we already interacted with her
							if( normalizedArousal >= 1.0f && normalizedArousal < 4.0f && lastInteractionDiff > 9.0f ) { // Now, we only bump up the arousal if we're under the maximum and if we let the girl rest for 9secs
								// We have a 25% chance to go to the next level of intensity
								int randBumpUp = UnityEngine.Random.Range(1, 5);
								if( bumpUpArousalAfterInactivity.val && randBumpUp == 4 ) {
									normalizedArousal += 1.0f;
									sexTriggerCount = Mathf.Round( normalizedArousal * ( maxTriggerCount.val / 4 ) );
								}
							}
						}
						
						currentArousal = normalizedArousal;
						lastInteractionTime = vammTimer;
					// Or we fall back to intensity 0
					} else {
						setArousal(0f);
					}
				}
				
				
				// Intensity boost (used for triggers for instance)
				float triggeredArousal = currentArousal;
				if( ArousalMode.val == "Manual" ) {
					if( currentArousal >= 0.0f && currentArousal < 4.0f ) {
						if( intensityBoost == true ) {
							triggeredArousal += 1.0f;
						}
					}
				}
				
				// Miscs temp vars
				AudioClip clip = null;

				// Randomized trigger calculation
				float randTriggerChance = -1f;
				bool canUseRandomTrigger = triggeredArousal >= 0.0f && triggeredArousal <= 4.0f;
				
				// We can only execute the random trigger if it is enabled and that we are in a situation where
				// it is a moan that is triggered, not breathing, kissing or blowjob
				if (enableRandMoanTrigger.val == true && canUseRandomTrigger == true)
				{
					moanLoopCount++;
					// We have reach the occurence where we wanna play a breathing sound instead of a moan
					if (moanLoopCount > moanRandOccurence)
					{
						// Calculating if we breath or moan
						randTriggerChance = UnityEngine.Random.Range(0f, 1f);
						// Resetting for the next occurence
						moanRandOccurence = (int) UnityEngine.Random.Range(randMoanTriggerOccurenceMin.val, randMoanTriggerOccurenceMax.val);
						moanLoopCount = 0;
					}
				}

				// Randomized trigger playback
				if (enableRandMoanTrigger.val == true && randTriggerChance >= 0f && randTriggerChance <= randMoanTriggerChance.val)
				{
					if (randMoanTriggerTypeChoice.val == "Breathing")
					{
						clip = currentClipPlaying = voice.GetTriggeredAudioBreath();
					}
					else if (randMoanTriggerTypeChoice.val == "Intensity 0" || randMoanTriggerTypeChoice.val == "Intensity 1" || randMoanTriggerTypeChoice.val == "Intensity 2" || randMoanTriggerTypeChoice.val == "Intensity 3" || randMoanTriggerTypeChoice.val == "Intensity 4")
					{
						string typeChoiceStr = randMoanTriggerTypeChoice.val;
						char[] separators = new char[] { ' ' };
						string[] typeChoiceArray = typeChoiceStr.Split(separators);
						float targetIntensity = float.Parse(typeChoiceArray[1]);
						clip = currentClipPlaying = voice.GetTriggeredAudioMoan(targetIntensity);
					}
					else if(randMoanTriggerTypeChoice.val == "Current Intensity -1")
					{
						float targetIntensity = UnityEngine.Mathf.Clamp(triggeredArousal - 1f, 0f, 3f);
						clip = currentClipPlaying = voice.GetTriggeredAudioMoan(targetIntensity);
					}
					else if(randMoanTriggerTypeChoice.val == "Current Intensity -2")
					{
						float targetIntensity = UnityEngine.Mathf.Clamp(triggeredArousal - 2f, 0f, 3f);
						clip = currentClipPlaying = voice.GetTriggeredAudioMoan(targetIntensity);
					}
					else if(randMoanTriggerTypeChoice.val == "Current Intensity +1")
					{
						float targetIntensity = UnityEngine.Mathf.Clamp(triggeredArousal + 1f, 1f, 4f);
						clip = currentClipPlaying = voice.GetTriggeredAudioMoan(targetIntensity);
					}
					else if(randMoanTriggerTypeChoice.val == "Current Intensity +2")
					{
						float targetIntensity = UnityEngine.Mathf.Clamp(triggeredArousal + 2f, 1f, 4f);
						clip = currentClipPlaying = voice.GetTriggeredAudioMoan(targetIntensity);
					}
					else if(randMoanTriggerTypeChoice.val == "Current Intensity +/-1")
					{
						float randIntensityMoan = UnityEngine.Random.Range(0, 2) * 2 - 1;
						float targetIntensity = UnityEngine.Mathf.Clamp(triggeredArousal + randIntensityMoan, 0f, 4f);
						clip = currentClipPlaying = voice.GetTriggeredAudioMoan(targetIntensity);
					}
					else if(randMoanTriggerTypeChoice.val == "Current Intensity +/-2")
					{
						float randIntensityMoan = ( UnityEngine.Random.Range(0, 2) * 2 - 1 ) * 2;
						float targetIntensity = UnityEngine.Mathf.Clamp(triggeredArousal + randIntensityMoan, 0f, 4f);
						clip = currentClipPlaying = voice.GetTriggeredAudioMoan(targetIntensity);
					}
					else if(randMoanTriggerTypeChoice.val == "Current Intensity +/-X")
					{
						float randIntensityMoan = ( UnityEngine.Random.Range(0, 2) * 2 - 1 ) * UnityEngine.Random.Range(1, 3);
						float targetIntensity = UnityEngine.Mathf.Clamp(triggeredArousal + randIntensityMoan, 0f, 4f);
						clip = currentClipPlaying = voice.GetTriggeredAudioMoan(targetIntensity);
					}
					


					if (clip != null)
					{
						float delayNextMoan = (clip.length / Math.Abs(VAMMPitch.val)) - 0.025f; // Overlapping the previous clip 				
						nextMoan = vammTimer + delayNextMoan;
						PlayAudio(clip);

						breathStart = vammTimer;
						breathEnd = nextMoan;
					}
				}
				// Or normal playback
				else
				{
					// Moans
					if (triggeredArousal >= 0.0f && triggeredArousal <= 4.0f)
					{
						clip = currentClipPlaying = voice.GetTriggeredAudioMoan(triggeredArousal);
						if (clip != null)
						{
							float delayBreathingEnd = (clip.length / Math.Abs(VAMMPitch.val)) + float.Parse(voice.voiceConfig["config"]["settings"]["overlapLengths"][(int) currentArousal].Value); // Overlapping the previous clip - this one is used for the breathing animation to be synced with the sample 				
							float delayNextMoan = delayBreathingEnd + VAMMPause.val; // Overlapping the previous clip + pause between clips
							
							delayBreathingEnd = vammTimer + delayBreathingEnd;
							nextMoan = vammTimer + delayNextMoan;
							
							PlayAudio(clip);

							breathStart = vammTimer;
							breathEnd = delayBreathingEnd;
						}
					// Breathing
					}
					else if (currentArousal == -1.0f)
					{
						clip = currentClipPlaying = voice.GetTriggeredAudioBreath();
						if (clip != null)
						{
							float delayNextMoan =
								(clip.length / Math.Abs(VAMMPitch.val)) - 0.025f; // Overlapping the previous clip 				
							nextMoan = vammTimer + delayNextMoan;
							PlayAudio(clip);

							breathStart = vammTimer;
							breathEnd = nextMoan;
						}
					// Kissing
					}
					else if (currentArousal == 100.0f)
					{
						clip = currentClipPlaying = voice.GetTriggeredAudioKiss();
						if (clip != null)
						{
							float delayNextMoan =
								(clip.length / Math.Abs(VAMMPitch.val)) + VAMMKissingSpeed.val -
								0.025f; // Overlapping the previous clip 				
							nextMoan = vammTimer + delayNextMoan;
							PlayAudio(clip);
						}
					// Blowjob
					}
					else if (currentArousal == 101.0f)
					{
						clip = currentClipPlaying = voice.GetTriggeredAudioBlowjob();
						if (clip != null)
						{
							float delayNextMoan = VAMMBlowjobSpeed.val; // Overlapping the previous clip 				
							nextMoan = vammTimer + delayNextMoan;
							PlayAudio(clip);
						}
					}
				}
            }
			
			// Controlling turn off while in interactive mode
			if( ArousalMode.val == "Interactive" ) {
				// We're only starting to turn off the girl after 10 seconds of non-interaction
				if( vammTimer >= nextTurnoffDecrease )
				{
					float lastCollisionDiff = vammTimer - lastCollisionTime;
					if( lastCollisionDiff > 10.0f ) {
						nextTurnoffDecrease = vammTimer + 1.0f;
						decreaseArousal();
					}
				}
			}
		}
		
		protected void setArousal( float arousal, bool ignoreTriggers=false ) {	
			// We don't change the arousal state if the character is in orgasm state
			// Or if the arousal you want is not in the list of those authorized in auto mode
			if( currentArousal != 10000.0f ) {
				// Executing triggers if we don't ignore them
				if( ignoreTriggers != true ) {
					executeTriggers( arousal );
				}
				
				// Changing current arousal
				currentArousal = arousal;
			}
        }
		
		protected float getInteractiveArousal(float valueToEvaluate = 0f) {
			float triggerCount = valueToEvaluate != null && valueToEvaluate > 0f ? valueToEvaluate : sexTriggerCount;
			float interactiveArousal = ( 4 * triggerCount / maxTriggerCount.val ) + 1;
			return interactiveArousal;
		}
		
		protected void decreaseArousal() {
			if( ArousalMode.val == "Interactive" ) {			
				sexTriggerCount = Mathf.Clamp( sexTriggerCount - sexTriggerDecayIncrement, 0.0f, maxTriggerCount.val );
			}
		}
		
		public string getArousalStateText() {
			// Bypassing the check between manual/interactive, because it is the normal arousal that drives the lock for orgasm
			if( currentArousal == 10000.0f ) return "Orgasming (locked until finished)";
			
			// Now checking depending on the mode
			int stateArousal = (int)(currentArousal);
			
			string stateText = "unknown";
			switch ( stateArousal ) {
				case -10:
					stateText = "Voice disabled";
					break;
				case -1:
					stateText = "Breathing";
					break;
				case 100:
					stateText = "Kissing";
					break;
				case 101:
					stateText = "Blowjob";
					break;
				case 0:
					stateText = "Intensity 0";
					break;
				case 1:
					stateText = "Intensity 1";
					break;
				case 2:
					stateText = "Intensity 2";
					break;
				case 3:
					stateText = "Intensity 3";
					break;
				case 4:
					stateText = "Intensity 4";
					break;
				case 5:
					stateText = "Intensity 5";
					break;
			}
			
			return stateText;
		}
				 
		public void ChangeVoice()
        {
			// We don't load anything if the scene is still loading
			if ( SuperController.singleton.isLoading ) {
				return;
			}
			
			// If we already have a voice in memory, unloading it
            if (voice != null)
            {
                voice.Unload();
            }
            voice = voices.GetVoice(selectedVoice);
            voice.Load();
			
			voiceInfos.val = "<color=#000><size=30><b>" + voice.name + ":</b></size></color>\n<color=#333><size=25>" + voice.infos + "</size></color>";			
			
			voiceInitialized = true;
        }
		
		public void PlayAudio(AudioClip clip)
		{
			if (clip != null)
			{
				headAudioSource.audioSource.PlayOneShot(clip);
			}
		}
	
		private string getStateInfos() {
			string stateVal = "<color=#000><size=30><b>State:</b></size></color>\n<color=#333><size=25>Character: <b>" + getArousalStateText() + "</b>\n";
			
			if( ArousalMode.val == "Interactive" ) {
				stateVal += "Arousal value: <b>" + (Mathf.Round(sexTriggerCount * 100f) / 100f) + "/" + maxTriggerCount.val + "</b>\n";
			}
			
			stateVal += "</size></color>";
			return stateVal;
		}

		private void updateRandomizedMoanHelp()
		{
			string moanCountStr = "";
			if (randMoanTriggerOccurenceMin.val != randMoanTriggerOccurenceMax.val)
			{
				moanCountStr = randMoanTriggerOccurenceMin.val + " to " + randMoanTriggerOccurenceMax.val;
			}
			else
			{
				moanCountStr = "" + randMoanTriggerOccurenceMin.val;
			}

			string helpText = "<color=#000><size=26><i><b>Enables randomized moan.</b>\nEvery " + moanCountStr + " moan the character will moan or play another type of sound instead. Playback chance is a percentage :\n " + (Mathf.Round(randMoanTriggerChance.val * 100f) / 100f) + " = " + Mathf.Round(randMoanTriggerChance.val*100) + "% chance to play another sample instead of moaning everytime it happens.</i></size></color>";
			randMoanTriggerHelp.val = helpText;
		}

		private void executeTriggers( float? targetArousal = null ) {
			
			// No triggers if we're not initialized
			if( !voiceInitialized ) return;
			
			// Checking triggers when we're switching arousal state
			if( targetArousal != null ) {
				// Trigger when disabling the voice
				if( targetArousal == -10f ) {
					startDisabledTrigger.Trigger();
					return; // We don't want to trigger anything else when disabling
				}
				
				// Triggers when lowering / increasing intensity, just between 0 and 4
				if( targetArousal >= 0f && targetArousal <= 4f && currentArousal >= 0f && currentArousal <= 4f ) {
					if( targetArousal < currentArousal ) {
						intensityLoweredTrigger.Trigger();
					} else if( targetArousal > currentArousal ) {
						intensityIncreasedTrigger.Trigger();
					}
				}
				
				// Checking our potential different cases
				if( targetArousal == -1.0f ) {
					startBreathingTrigger.Trigger();
				} else if( targetArousal == 0f ) {
					startIntensity0Trigger.Trigger();
				} else if( targetArousal == 1f ) {
					startIntensity1Trigger.Trigger();
				} else if( targetArousal == 2f ) {
					startIntensity2Trigger.Trigger();
				} else if( targetArousal == 3f ) {
					startIntensity3Trigger.Trigger();
				} else if( targetArousal == 4f ) {
					startIntensity4Trigger.Trigger();
				} else if( targetArousal == 5f ) {
					reachOrgasmTrigger.Trigger();
				} else if( targetArousal == 100.0f ) {
					startKissingTrigger.Trigger();
				} else if( targetArousal == 101.0f ) {
					startBlowjobTrigger.Trigger();
				}
			}
		}
		
		// ***************************************
		// Everything related to triggers
		// ***************************************	
		private void initTriggers() {
			// Creating all my triggers and adding them to a global list (to easily control them and save them)
			allTriggers = new List<EventTrigger>();
			
			startDisabledTrigger = new EventTrigger(this, "StartDisabledActions");
			allTriggers.Add(startDisabledTrigger);
			
			startBreathingTrigger = new EventTrigger(this, "StartBreathingActions");
			allTriggers.Add(startBreathingTrigger);
			
			startKissingTrigger = new EventTrigger(this, "StartKissingActions");
			allTriggers.Add(startKissingTrigger);
			
			startBlowjobTrigger = new EventTrigger(this, "StartBlowjobActions");
			allTriggers.Add(startBlowjobTrigger);
			
			startIntensity0Trigger = new EventTrigger(this, "StartIntensity0Actions");
			allTriggers.Add(startIntensity0Trigger);
			
			startIntensity1Trigger = new EventTrigger(this, "StartIntensity1Actions");
			allTriggers.Add(startIntensity1Trigger);
			
			startIntensity2Trigger = new EventTrigger(this, "StartIntensity2Actions");
			allTriggers.Add(startIntensity2Trigger);
			
			startIntensity3Trigger = new EventTrigger(this, "StartIntensity3Actions");
			allTriggers.Add(startIntensity3Trigger);
			
			startIntensity4Trigger = new EventTrigger(this, "StartIntensity4Actions");
			allTriggers.Add(startIntensity4Trigger);
			
			intensityLoweredTrigger = new EventTrigger(this, "IntensityLoweredActions");
			allTriggers.Add(intensityLoweredTrigger);
			
			intensityIncreasedTrigger = new EventTrigger(this, "IntensityIncreasedActions");
			allTriggers.Add(intensityIncreasedTrigger);
			
			reachOrgasmTrigger = new EventTrigger(this, "ReachOrgasmActions");
			allTriggers.Add(reachOrgasmTrigger);
			
			endOrgasmTrigger = new EventTrigger(this, "EndOrgasmActions");
			allTriggers.Add(endOrgasmTrigger);
			
			breathingTrigger = new EventTrigger(this, "BreathingActions");
			allTriggers.Add(breathingTrigger);
		
			// Loading triggers assets
			SimpleTriggerHandler.LoadAssets();
			
			// Adding the rename listener
			SuperController.singleton.onAtomUIDRenameHandlers += OnAtomRename;
		}
		
		private void OnAtomRename(string oldid, string newid)
		{
			foreach( EventTrigger et in allTriggers ) {
				et.SyncAtomNames();
			}
		}
		
		// ***************************************
		// Everything related to morphs animation
		// ***************************************		
		private void updateMorphs() {
			if( VAMMBreathingEnabled.val == false ) return;
			
			if( breathStart != null && breathEnd != null ) {
				if( ( currentArousal >= -1.0f && currentArousal <= 4.0f ) || currentArousal == 5.0f || currentArousal == 10000.0f ) {
					if( lifeChestMorph != null && currentClipPlaying != null ) {
						
						// Selecting my morphs max values
						float chestMorphStartValue;
						float chestMorphEndValue;
						float stomachMorphStartValue;
						float stomachMorphEndValue;
						float noseMorphStartValue;
						float noseMorphEndValue;
						
						float breathChestStartValue = 0.0f;
						float breathChestTargetValue = -0.45f;
						float breathStomachStartValue = 0.0f;
						float breathStomachTargetValue = -0.30f;
						float noseStartValue = 0.40f;
						float noseTargetValue = 0f;
						
						if( currentArousal == 3.0f ) {
							breathChestTargetValue = -0.35f;
							breathStomachTargetValue = -0.25f;
						} else if(currentArousal == 4.0f ) {
							breathChestTargetValue = -0.25f;
							breathStomachTargetValue = -0.20f;
						}
						
						// Selecting curves
						AnimationCurve currentCurveCh = ACBreathingCh;
						AnimationCurve currentCurveSt = ACBreathingSt;

						// Chest
						if( currentArousal == 4.0f ) {
							currentCurveCh = ACIntensity4Ch;
						} else if( currentArousal == 5.0f || currentArousal == 10000.0f ) {
							currentCurveCh = ACOrgasmCh;
						}
						
						// Stomach
						if( currentArousal >= 1.0f ) {
							currentCurveSt = ACIntensity1St;
						}
						
						// Settings my final vars modified by the breathing scale and doing the interpolation
						chestMorphStartValue = breathChestStartValue * VAMMBreathingScale.val;
						chestMorphEndValue = breathChestTargetValue * VAMMBreathingScale.val;
						stomachMorphStartValue = breathStomachStartValue * VAMMBreathingScale.val;
						stomachMorphEndValue = breathStomachTargetValue * VAMMBreathingScale.val;
						noseMorphStartValue = noseStartValue * VAMMBreathingScale.val;
						noseMorphEndValue = noseTargetValue * VAMMBreathingScale.val;

						breathPeriod = (vammTimer-breathStart)/(breathEnd-breathStart);
						
						lifeChestMorph.morphValue = Mathf.Lerp(chestMorphStartValue, chestMorphEndValue, currentCurveCh.Evaluate(breathPeriod) );
						lifeStomachMorph.morphValue = Mathf.Lerp(stomachMorphStartValue, stomachMorphEndValue, currentCurveSt.Evaluate(breathPeriod) );
						lifeNoseOutMorph.morphValue = Mathf.Lerp(noseMorphStartValue, noseMorphEndValue, currentCurveSt.Evaluate(breathPeriod) );
						breathingTrigger.active = true;
						breathingTrigger.TriggerTransition( currentCurveCh.Evaluate(breathPeriod) );
					}
				}
			}
		}
		
		private void resetMorphs() {
			if( lifeChestMorph != null ) {
				lifeChestMorph.morphValue = 0.0f;
				lifeStomachMorph.morphValue = 0.0f;
				lifeNoseOutMorph.morphValue = 0.0f;
			}
		}
			
		// **************************
		// Callbacks and similar
		// **************************
		protected void switchArousalModeCallback(string selectedmode) {
			// Initializing if we're switching values
			// Breathing
			if( selectedmode == "Manual" ) {
				currentArousal = -1.0f; 
			} else {
				currentArousal = 0; 
			}
			// Resetting trigger count
			sexTriggerCount = 0;
			
		}
	    
		protected void maxTriggerCountCallback ( float triggerCount ) {
			if (VAMMValue_MaxTriggerCount != null)
			{
				VAMMValue_MaxTriggerCount.val = triggerCount;
			}
		}
		
		protected void arousalRateCallback( float arousalRate ) {}
		
		protected void VAMMVolumeCallback ( float volumeval ) {
			headAudioSource.SetFloatParamValue("volume",volumeval);
		}
		
		protected void VAMMPitchCallback ( float pitchval ) {
			headAudioSource.SetFloatParamValue("pitch",pitchval);
		}
		
		protected void VAMMPauseCallback ( float pitchval ) {
			// Nothing for now
		}

		protected void VAMMReverbEnabledCallback( bool spatializeval ) {
			headAudioSource.SetBoolParamValue("spatialize", !spatializeval );
		}
		
		protected void VAMMBreathingEnabledCallback( bool state ) {
			
		}

		protected void randMoanTriggerOccurenceMinCallback(float min)
		{
			updateRandomizedMoanHelp();
		}
		
		protected void randMoanTriggerOccurenceMaxCallback(float min)
		{
			updateRandomizedMoanHelp();
		}
		
		protected void randMoanTriggerChanceCallback(float scale)
		{
			updateRandomizedMoanHelp();
		}
		
		protected void VAMMBreathingScaleCallback( float scale ) {
			resetMorphs();
		}
		
		protected void VAMMAutoJawCallback( bool state ) {
			JawControlAJ.SetBoolParamValue("driveXRotationFromAudioSource", state); // Setting the jaw control in "Misc Physics"
			AutoJawMouthMorphsJS.SetBoolParamValue("enabled", state); // Settings the jaw mouth control in "Auto behaviors"
		}
		
		protected void setOptimalJawParamsCallback() {
			if( voice != null ) {
				JawControlAJ.SetFloatParamValue("spring", float.Parse( voice.voiceConfig["config"]["settings"]["jawHoldSpring"].Value ) );
				JawControlAJ.SetFloatParamValue("damper", float.Parse( voice.voiceConfig["config"]["settings"]["jawHoldDamper"].Value ) );
				JawControlAJ.SetFloatParamValue("driveXRotationFromAudioSourceMultiplier", float.Parse( voice.voiceConfig["config"]["settings"]["audioDriveMultiplier"].Value ) );
				JawControlAJ.SetFloatParamValue("driveXRotationFromAudioSourceAdditionalAngle", float.Parse( voice.voiceConfig["config"]["settings"]["audioDriveExtraAngle"].Value ) );
				JawControlAJ.SetFloatParamValue("driveXRotationFromAudioSourceMaxAngle", float.Parse( voice.voiceConfig["config"]["settings"]["audioDriveMaxAngle"].Value ) );
			}
		}

			
		// **************************
		// Collision events
		// **************************	
		protected void VaTrigObserver(object sender, TriggerEventArgs e)
        {
			if( ArousalMode.val != "Interactive" ) return;
			if (e.evtType == "Entered" && vammTimer >= nextTriggerAllowed )
			{
				float normalizedArousal = (float)Math.Floor( getInteractiveArousal( sexTriggerCount + 1 ) );
				executeTriggers( normalizedArousal );
				
				sexTriggerCount += 1.0f * arousalRate.val;
				
				// Updating the last collision time
				lastCollisionTime = vammTimer;
			}
        }
		
		// **************************
		// OVERRIDES FOR THE TRIGGERS
		// **************************	
		public override JSONClass GetJSON(bool includePhysical = true, bool includeAppearance = true, bool forceStore = false)
		{
			JSONClass jc = base.GetJSON(includePhysical, includeAppearance, forceStore);
			if (includePhysical || forceStore)
			{
				needsStore = true;
				foreach( EventTrigger et in allTriggers ) {
					jc[et.Name] = et.GetJSON(base.subScenePrefix);
				}
				
			}
			return jc;
		}

		public override void LateRestoreFromJSON(JSONClass jc, bool restorePhysical = true, bool restoreAppearance = true, bool setMissingToDefault = true)
		{
			base.LateRestoreFromJSON(jc, restorePhysical, restoreAppearance, setMissingToDefault);
			if (!base.physicalLocked && restorePhysical && !IsCustomPhysicalParamLocked("trigger"))
			{
				foreach( EventTrigger et in allTriggers ) {
					et.RestoreFromJSON(jc, base.subScenePrefix, base.mergeRestore, setMissingToDefault);
				}
			}
		}
		
		// **************************
		// MY TOOLS
		// **************************
		private static void logDebug( string debugText ) {
			SuperController.LogMessage( debugText );
		}
		
		private static void disableScrollOnText(UIDynamicTextField target) {
			ScrollRect targetSR = target.UItext.transform.parent.transform.parent.transform.parent.GetComponent<ScrollRect>();
			if( targetSR != null ) {
				targetSR.horizontal = false;
				targetSR.vertical = false;
			}
		}
		
		private JSONStorableString createStaticDescriptionText( string DescTitle, string DescText, bool rightSide, int fieldHeight ) {
			JSONStorableString staticDescString = new JSONStorableString(DescTitle,DescText);
			UIDynamicTextField staticDescStringField = CreateTextField(staticDescString, rightSide);
			staticDescStringField.backgroundColor = new Color(1f, 1f, 1f, 0f);
			LayoutElement sdsfLayout = staticDescStringField.GetComponent<LayoutElement>();
			sdsfLayout.preferredHeight = sdsfLayout.minHeight = fieldHeight;
			staticDescStringField.height = fieldHeight;
			disableScrollOnText(staticDescStringField);

			return staticDescString;
		}
		
		// ***********************************************************
		// EXTERNAL TOOLS - Thank you great coders for your content!
		// ***********************************************************
		
		// *********** MacGruber_Utils.cs START *********************
		// Get directory path where the plugin is located. Based on Alazi's & VAMDeluxe's method.
		public static string GetPluginPath(MVRScript self)
		{
			string id = self.name.Substring(0, self.name.IndexOf('_'));
			string filename = self.manager.GetJSON()["plugins"][id].Value;
			return filename.Substring(0, filename.LastIndexOfAny(new char[] { '/', '\\' }));
		}
				
		// Get path prefix of the package that contains our plugin.
		public static string GetPackagePath(MVRScript self)
		{
			string id = self.name.Substring(0, self.name.IndexOf('_'));
			string filename = self.manager.GetJSON()["plugins"][id].Value;
			int idx = filename.IndexOf(":/");
			if (idx >= 0)
				return filename.Substring(0, idx+2);
			else
				return string.Empty;
		}
				
		// Check if our plugin is running from inside a package
		public static bool IsInPackage(MVRScript self)
		{
			string id = self.name.Substring(0, self.name.IndexOf('_'));
			string filename = self.manager.GetJSON()["plugins"][id].Value;
			return filename.IndexOf(":/") >= 0;
		}	
		
		// ===========================================================================================
		// TriggerHandler implementation for easier handling of custom triggers.
		// Essentially call this in your plugin init code:
		//     StartCoroutine(SimpleTriggerHandler.LoadAssets());
		//
		// Credit to AcidBubbles for figuring out how to do custom triggers.
		//
		// hazmhox note : Edited a bit to allow the use of transitions ( original code coming from LogicBricks 4 )
		public class SimpleTriggerHandler : TriggerHandler
		{
			public static bool Loaded { get; private set; }

			private static SimpleTriggerHandler myInstance;

			private RectTransform myTriggerActionsPrefab;
			private RectTransform myTriggerActionMiniPrefab;
			private RectTransform myTriggerActionDiscretePrefab;
			private RectTransform myTriggerActionTransitionPrefab;

			public static SimpleTriggerHandler Instance {
				get {
					if (myInstance == null)
						myInstance = new SimpleTriggerHandler();
					return myInstance;
				}
			}

			public static void LoadAssets()
			{
				SuperController.singleton.StartCoroutine(Instance.LoadAssetsInternal());
			}

			private IEnumerator LoadAssetsInternal()
			{
				foreach (var x in LoadAsset("z_ui2", "TriggerActionsPanel", p => myTriggerActionsPrefab = p))
					yield return x;
				foreach (var x in LoadAsset("z_ui2", "TriggerActionMiniPanel", p => myTriggerActionMiniPrefab = p))
					yield return x;
				foreach (var x in LoadAsset("z_ui2", "TriggerActionDiscretePanel", p => myTriggerActionDiscretePrefab = p))
					yield return x;
				foreach (var x in LoadAsset("z_ui2", "TriggerActionTransitionPanel", p => myTriggerActionTransitionPrefab = p))
					yield return x;

				Loaded = true;
			}

			private IEnumerable LoadAsset(string assetBundleName, string assetName, Action<RectTransform> assign)
			{
				AssetBundleLoadAssetOperation request = AssetBundleManager.LoadAssetAsync(assetBundleName, assetName, typeof(GameObject));
				if (request == null)
					throw new NullReferenceException($"Request for {assetName} in {assetBundleName} assetbundle failed: Null request.");
				yield return request;
				GameObject go = request.GetAsset<GameObject>();
				if (go == null)
					throw new NullReferenceException($"Request for {assetName} in {assetBundleName} assetbundle failed: Null GameObject.");
				RectTransform prefab = go.GetComponent<RectTransform>();
				if (prefab == null)
					throw new NullReferenceException($"Request for {assetName} in {assetBundleName} assetbundle failed: Null RectTansform.");
				assign(prefab);
			}


			void TriggerHandler.RemoveTrigger(Trigger t)
			{
				// nothing to do
			}

			void TriggerHandler.DuplicateTrigger(Trigger t)
			{
				throw new NotImplementedException();
			}

			RectTransform TriggerHandler.CreateTriggerActionsUI()
			{
				return UnityEngine.Object.Instantiate(myTriggerActionsPrefab);
			}

			RectTransform TriggerHandler.CreateTriggerActionMiniUI()
			{
				return UnityEngine.Object.Instantiate(myTriggerActionMiniPrefab);
			}

			RectTransform TriggerHandler.CreateTriggerActionDiscreteUI()
			{
				return UnityEngine.Object.Instantiate(myTriggerActionDiscretePrefab);
			}

			RectTransform TriggerHandler.CreateTriggerActionTransitionUI()
			{
				return UnityEngine.Object.Instantiate(myTriggerActionTransitionPrefab);
			}

			void TriggerHandler.RemoveTriggerActionUI(RectTransform rt)
			{
				UnityEngine.Object.Destroy(rt?.gameObject);
			}
		}

		// Wrapper for easier handling of custom triggers.
		public class EventTrigger : Trigger
		{
			public string Name {
				get { return name; }
				set { name = value; myNeedInit = true; }
			}

			public MVRScript Owner {
				get; private set;
			}

			private string name;
			private bool myNeedInit = true;

			public EventTrigger(MVRScript owner, string name)
			{
				Name = name;
				Owner = owner;
				handler = SimpleTriggerHandler.Instance;
			}

			public void OpenPanelActionStart()
			{
				if (!SimpleTriggerHandler.Loaded)
				{
					SuperController.LogError("EventTrigger: You need to call SimpleTriggerHandler.LoadAssets() before use.");
					return;
				}

				triggerActionsParent = Owner.UITransform;
				InitTriggerUI();
				OpenTriggerActionsPanel();
				if (myNeedInit)
				{
					triggerActionsPanel.Find("Panel/Header Text").GetComponent<Text>().text = Name;
					triggerActionsPanel.Find("Content/Tab1/Label").GetComponent<Text>().text = "Event Actions";
					triggerActionsPanel.Find("Content/Tab2").gameObject.SetActive(false);
					triggerActionsPanel.Find("Content/Tab3").gameObject.SetActive(false);
					myNeedInit = false;
				}
			}
			
			public void OpenPanelTransition()
			{
				if (!SimpleTriggerHandler.Loaded)
				{
					SuperController.LogError("EventTrigger: You need to call SimpleTriggerHandler.LoadAssets() before use.");
					return;
				}

				triggerActionsParent = Owner.UITransform;
				InitTriggerUI();
				OpenTriggerActionsPanel();
				
				//getRoot(triggerActionsParent,0);
				if (myNeedInit)
				{
					triggerActionsPanel.Find("Panel/Header Text").GetComponent<Text>().text = Name;
					triggerActionsPanel.Find("Content/Tab2/Label").GetComponent<Text>().text = "Breathing Actions";
					triggerActionsPanel.Find("Content/Tab1").gameObject.SetActive(false);
					triggerActionsPanel.Find("Content/Tab2").gameObject.SetActive(true);
					triggerActionsPanel.Find("Content/Tab2").GetComponent<Toggle>().isOn = true;					
					triggerActionsPanel.Find("Content/Tab3").gameObject.SetActive(false);
					myNeedInit = false;
				}
			}
	
			public void RestoreFromJSON(JSONClass jc, string subScenePrefix, bool isMerge, bool setMissingToDefault)
			{
				if (jc.HasKey(Name))
				{
					JSONClass tc = jc[Name].AsObject;
					if (tc != null)
						base.RestoreFromJSON(tc, subScenePrefix, isMerge);
				}
				else if (setMissingToDefault)
				{
					base.RestoreFromJSON(new JSONClass());
				}
			}

			public void Trigger()
			{
				active = true;
				active = false;
			}
			
			public void TriggerTransition(float val)
			{
				this.transitionInterpValue = val;
			}
		}
		// *********** MacGruber_Utils.cs END *********************
		
		// **************************
		// Time to cleanup !
		// **************************
		void OnDisable() {
			restoreAttenuation();
		}
		
		void OnDestroy() {
			// All event triggers
			SuperController.singleton.onAtomUIDRenameHandlers -= OnAtomRename;
			foreach( EventTrigger et in allTriggers ) {
				et.Remove();
			}
			
			// Collision triggers / events
			vaTrig.gameObject.GetComponent<SexTriggerCollide>().OnCollide -= VaTrigObserver;
			Destroy( vaTrig.gameObject.AddComponent<SexTriggerCollide>() );

			// Sounds
			if (voices != null)
            {
                voices.UnloadAudio();
            }
		}
	}
}