using UnityEngine;
using System;
using System.Collections;
using System.Collections.Generic;
using SimpleJSON;
using System.Linq;
using System.IO;

// VAMMoanAFX v1.0
//
// Audio effects for VAMMoan

namespace VAMMoanAFXPlugin
{
    public class VAMMoanAFX : MVRScript
    {
		
        // Audio Listener components
		protected AudioEchoFilter AFXEchoFilter;
		protected AudioReverbFilter AFXReverbFilter;
		protected AudioChorusFilter AFXChorusFilter;
		protected AudioLowPassFilter AFXLPFilter;
		protected AudioHighPassFilter AFXHPFilter;
		
		// Variables for the effects
		// Reverb
		public JSONStorableBool RevEnabled;
		public JSONStorableFloat RevDecayTime;
		public JSONStorableFloat RevDecayHFRatio;
		public JSONStorableFloat RevDensity;
		public JSONStorableFloat RevDiffusion;
		public JSONStorableFloat RevDryLevel;
					
		// Echo
		public JSONStorableBool EchEnabled;
		public JSONStorableFloat EchDecayRatio;
		public JSONStorableFloat EchDelay;
		public JSONStorableFloat EchDryMix;
		public JSONStorableFloat EchWetMix;
		
		// Chorus
		public JSONStorableBool ChoEnabled;
		public JSONStorableFloat ChoDelay;
		public JSONStorableFloat ChoDepth;
		public JSONStorableFloat ChoDryMix;
		public JSONStorableFloat ChoRate;
		public JSONStorableFloat ChoWetMix1;
		public JSONStorableFloat ChoWetMix2;
		public JSONStorableFloat ChoWetMix3;
		
		// Low Pass
		public JSONStorableBool LPEnabled;
		public JSONStorableFloat LPCutoffFreq;
		public JSONStorableFloat LPResonance;

		// High Pass
		public JSONStorableBool HPEnabled;
		public JSONStorableFloat HPCutoffFreq;
		public JSONStorableFloat HPResonance;
		
		public override void Init()
        {			 
            try
            {
                if (containingAtom.type != "Person")
                {
                    SuperController.LogError("Please add VAMMoanAFX on an Person Atom");
                    return;
                }
								
				if( containingAtom.type == "Person" ) {
					InitCurrentAudioSource();
				}
				
				// Merge load fix / preventing the plugin to load if it hasn't been initialized properly
				if( AFXReverbFilter == null ) {
					return;
				}
				
				// *****************************
				// REVERB
				// *****************************

				JSONStorableString reverbTitle = new JSONStorableString("Reverb",
					"<color=#000><size=55><b>Reverb</b></size></color>\n" +
					"<color=#333><size=22><b>Simulate how the sound is reverberated depending on the size of the room and its surroundings</b></size></color>\n"
				);
				UIDynamic reverbTitlefield = CreateTextField(reverbTitle, false);
				reverbTitlefield.height = 180.0f;
				
				RevEnabled = new JSONStorableBool("Reverb enabled", false, ( val ) => { AFXReverbFilter.enabled = val; });
				RevEnabled.storeType = JSONStorableParam.StoreType.Full;
				CreateToggle(RevEnabled, false);
				RegisterBool(RevEnabled);
				
				RevDecayTime = new JSONStorableFloat("Decay Time", 1.0f, ( val ) => { AFXReverbFilter.decayTime = val; }, 0.1f, 20.0f, true, true);
				RevDecayTime.storeType = JSONStorableParam.StoreType.Full;
				CreateSlider(RevDecayTime, false);
				RegisterFloat(RevDecayTime);
				
				RevDecayHFRatio = new JSONStorableFloat("H.F. decay time ratio", 0.5f, ( val ) => { AFXReverbFilter.decayHFRatio = val; }, 0.1f, 2.0f, true, true);
				RevDecayHFRatio.storeType = JSONStorableParam.StoreType.Full;
				CreateSlider(RevDecayHFRatio, false);
				RegisterFloat(RevDecayHFRatio);
				
				RevDensity = new JSONStorableFloat("Density", 100.0f, ( val ) => { AFXReverbFilter.density = val; }, 0.0f, 100.0f, true, true);
				RevDensity.storeType = JSONStorableParam.StoreType.Full;
				CreateSlider(RevDensity, false);
				RegisterFloat(RevDensity);
				
				RevDiffusion = new JSONStorableFloat("Diffusion", 100.0f, ( val ) => { AFXReverbFilter.diffusion = val; }, 0.0f, 100.0f, true, true);
				RevDiffusion.storeType = JSONStorableParam.StoreType.Full;
				CreateSlider(RevDiffusion, false);
				RegisterFloat(RevDiffusion);
				
				RevDryLevel = new JSONStorableFloat("Dry Level", 0.0f, ( val ) => { AFXReverbFilter.dryLevel = val; }, -10000.0f, 0.0f, true, true);
				RevDryLevel.storeType = JSONStorableParam.StoreType.Full;
				CreateSlider(RevDryLevel, false);
				RegisterFloat(RevDryLevel);

				
				// *****************************
				// CHORUS
				// *****************************
				JSONStorableString chorusTitle = new JSONStorableString("Chorus",
					"<color=#000><size=55><b>Chorus</b></size></color>\n" +
					"<color=#333><size=22><b>Adds a chorus effect to the sound (like if the sound was playing several times slightly delayed)</b></size></color>\n"
				);
				UIDynamic chorusTitlefield = CreateTextField(chorusTitle, false);
				chorusTitlefield.height = 180.0f;
				
				ChoEnabled = new JSONStorableBool("Chorus enabled", false, ( val ) => { AFXChorusFilter.enabled = val; });
				ChoEnabled.storeType = JSONStorableParam.StoreType.Full;
				CreateToggle(ChoEnabled, false);
				RegisterBool(ChoEnabled);
				
				ChoDelay = new JSONStorableFloat("Decay ratio", 40.0f, ( val ) => { AFXChorusFilter.delay = val; }, 0.1f, 100.0f, true, true);
				ChoDelay.storeType = JSONStorableParam.StoreType.Full;
				CreateSlider(ChoDelay, false);
				RegisterFloat(ChoDelay);
				
				ChoDepth = new JSONStorableFloat("Delay", 0.03f, ( val ) => { AFXChorusFilter.depth = val; }, 0.0f, 1.0f, true, true);
				ChoDepth.storeType = JSONStorableParam.StoreType.Full;
				CreateSlider(ChoDepth, false);
				RegisterFloat(ChoDepth);
				
				ChoDryMix = new JSONStorableFloat("Dry mix", 0.5f, ( val ) => { AFXChorusFilter.dryMix = val; }, 0.0f, 1.0f, true, true);
				ChoDryMix.storeType = JSONStorableParam.StoreType.Full;
				CreateSlider(ChoDryMix, false);
				RegisterFloat(ChoDryMix);
				
				ChoRate = new JSONStorableFloat("Rate", 0.8f, ( val ) => { AFXChorusFilter.rate = val; }, 0.0f, 20.0f, true, true);
				ChoRate.storeType = JSONStorableParam.StoreType.Full;
				CreateSlider(ChoRate, false);
				RegisterFloat(ChoRate);
				
				ChoWetMix1 = new JSONStorableFloat("Wet mix 1", 0.5f, ( val ) => { AFXChorusFilter.wetMix1 = val; }, 0.0f, 1.0f, true, true);
				ChoWetMix1.storeType = JSONStorableParam.StoreType.Full;
				CreateSlider(ChoWetMix1, false);
				RegisterFloat(ChoWetMix1);
				
				ChoWetMix2 = new JSONStorableFloat("Wet mix 2", 0.5f, ( val ) => { AFXChorusFilter.wetMix2 = val; }, 0.0f, 1.0f, true, true);
				ChoWetMix2.storeType = JSONStorableParam.StoreType.Full;
				CreateSlider(ChoWetMix2, false);
				RegisterFloat(ChoWetMix2);
				
				ChoWetMix3 = new JSONStorableFloat("Wet mix 3", 0.5f, ( val ) => { AFXChorusFilter.wetMix3 = val; }, 0.0f, 1.0f, true, true);
				ChoWetMix3.storeType = JSONStorableParam.StoreType.Full;
				CreateSlider(ChoWetMix3, false);
				RegisterFloat(ChoWetMix3);
				
				
				// *****************************
				// ECHO
				// *****************************				
				JSONStorableString echoTitle = new JSONStorableString("Echo",
					"<color=#000><size=55><b>Echo</b></size></color>\n" +
					"<color=#333><size=22><b>Simulate how the sound echoes on the surroundings</b></size></color>\n"
				);
				UIDynamic echoTitlefield = CreateTextField(echoTitle, true);
				echoTitlefield.height = 160.0f;
				
				EchEnabled = new JSONStorableBool("Echo enabled", false, ( val ) => { AFXEchoFilter.enabled = val; });
				EchEnabled.storeType = JSONStorableParam.StoreType.Full;
				CreateToggle(EchEnabled, true);
				RegisterBool(EchEnabled);
				
				EchDecayRatio = new JSONStorableFloat("Decay ratio", 0.5f, ( val ) => { AFXEchoFilter.decayRatio = val; }, 0.0f, 1.0f, true, true);
				EchDecayRatio.storeType = JSONStorableParam.StoreType.Full;
				CreateSlider(EchDecayRatio, true);
				RegisterFloat(EchDecayRatio);
				
				EchDelay = new JSONStorableFloat("Delay", 500.0f, ( val ) => { AFXEchoFilter.delay = val; }, 10.0f, 5000.0f, true, true);
				EchDelay.storeType = JSONStorableParam.StoreType.Full;
				CreateSlider(EchDelay, true);
				RegisterFloat(EchDelay);
				
				EchDryMix = new JSONStorableFloat("Dry mix", 1.0f, ( val ) => { AFXEchoFilter.dryMix = val; }, 0.0f, 1.0f, true, true);
				EchDryMix.storeType = JSONStorableParam.StoreType.Full;
				CreateSlider(EchDryMix, true);
				RegisterFloat(EchDryMix);
				
				EchWetMix = new JSONStorableFloat("Wet mix", 1.0f, ( val ) => { AFXEchoFilter.wetMix = val; }, 0.0f, 1.0f, true, true);
				EchWetMix.storeType = JSONStorableParam.StoreType.Full;
				CreateSlider(EchWetMix, true);
				RegisterFloat(EchWetMix);
				
				// *****************************
				// Low Pass
				// *****************************
				JSONStorableString lpTitle = new JSONStorableString("Low pass",
					"<color=#000><size=55><b>Low pass</b></size></color>\n" +
					"<color=#333><size=22><b>Filter high frequencies (treble) to only allow low frequencies (bass)</b></size></color>\n"
				);
				UIDynamic lpTitlefield = CreateTextField(lpTitle, true);
				lpTitlefield.height = 160.0f;
				
				LPEnabled = new JSONStorableBool("Low pass enabled", false, ( val ) => { AFXLPFilter.enabled = val; });
				LPEnabled.storeType = JSONStorableParam.StoreType.Full;
				CreateToggle(LPEnabled, true);
				RegisterBool(LPEnabled);
				
				LPCutoffFreq = new JSONStorableFloat("LP Cutoff frequency", 5000.0f, ( val ) => { AFXLPFilter.cutoffFrequency = val; }, 10.0f, 22000.0f, true, true);
				LPCutoffFreq.storeType = JSONStorableParam.StoreType.Full;
				CreateSlider(LPCutoffFreq, true);
				RegisterFloat(LPCutoffFreq);
				
				LPResonance = new JSONStorableFloat("Low pass resonance", 1.0f, ( val ) => { AFXLPFilter.lowpassResonanceQ = val; }, 1.0f, 10.0f, true, true);
				LPResonance.storeType = JSONStorableParam.StoreType.Full;
				CreateSlider(LPResonance, true);
				RegisterFloat(LPResonance);
				
				
				// *****************************
				// High Pass
				// *****************************
				JSONStorableString hpTitle = new JSONStorableString("High pass",
					"<color=#000><size=55><b>High pass</b></size></color>\n" +
					"<color=#333><size=22><b>Filter low frequencies (bass) to only allow high frequencies (treble)</b></size></color>\n"
				);
				UIDynamic hpTitlefield = CreateTextField(hpTitle, true);
				hpTitlefield.height = 160.0f;
				
				HPEnabled = new JSONStorableBool("High pass enabled", false, ( val ) => { AFXHPFilter.enabled = val; });
				HPEnabled.storeType = JSONStorableParam.StoreType.Full;
				CreateToggle(HPEnabled, true);
				RegisterBool(HPEnabled);
				
				HPCutoffFreq = new JSONStorableFloat("HP Cutoff frequency", 5000.0f, ( val ) => { AFXHPFilter.cutoffFrequency = val; }, 10.0f, 22000.0f, true, true);
				HPCutoffFreq.storeType = JSONStorableParam.StoreType.Full;
				CreateSlider(HPCutoffFreq, true);
				RegisterFloat(HPCutoffFreq);
				
				HPResonance = new JSONStorableFloat("High pass resonance", 1.0f, ( val ) => { AFXHPFilter.highpassResonanceQ = val; }, 1.0f, 10.0f, true, true);
				HPResonance.storeType = JSONStorableParam.StoreType.Full;
				CreateSlider(HPResonance, true);
				RegisterFloat(HPResonance);
				
				
				// User interface is ok, now let's init the values properly
				InitAFXValues();
				

			}
            catch(Exception e)
            {
                SuperController.LogError("VAMMoanAFX - Exception caught: " + e);
            }
		}
		
		void Start()
        {
			
        }
			
		void InitCurrentAudioSource() {
			AFXEchoFilter = containingAtom.GetStorableByID("HeadAudioSource").gameObject.GetComponent<AudioEchoFilter>();
			if( AFXEchoFilter == null ) {
				containingAtom.GetStorableByID("HeadAudioSource").gameObject.AddComponent<AudioEchoFilter>();
				AFXEchoFilter = containingAtom.GetStorableByID("HeadAudioSource").gameObject.GetComponent<AudioEchoFilter>();
			}
			
			AFXReverbFilter = containingAtom.GetStorableByID("HeadAudioSource").gameObject.GetComponent<AudioReverbFilter>();
			if( AFXReverbFilter == null ) {
				containingAtom.GetStorableByID("HeadAudioSource").gameObject.AddComponent<AudioReverbFilter>();
				AFXReverbFilter = containingAtom.GetStorableByID("HeadAudioSource").gameObject.GetComponent<AudioReverbFilter>();
			}
			
			AFXChorusFilter = containingAtom.GetStorableByID("HeadAudioSource").gameObject.GetComponent<AudioChorusFilter>();
			if( AFXChorusFilter == null ) {
				containingAtom.GetStorableByID("HeadAudioSource").gameObject.AddComponent<AudioChorusFilter>();
				AFXChorusFilter = containingAtom.GetStorableByID("HeadAudioSource").gameObject.GetComponent<AudioChorusFilter>();
			}
			
			AFXLPFilter = containingAtom.GetStorableByID("HeadAudioSource").gameObject.GetComponent<AudioLowPassFilter>();
			if( AFXLPFilter == null ) {
				containingAtom.GetStorableByID("HeadAudioSource").gameObject.AddComponent<AudioLowPassFilter>();
				AFXLPFilter = containingAtom.GetStorableByID("HeadAudioSource").gameObject.GetComponent<AudioLowPassFilter>();
			}
			
			AFXHPFilter = containingAtom.GetStorableByID("HeadAudioSource").gameObject.GetComponent<AudioHighPassFilter>();
			if( AFXHPFilter == null ) {
				containingAtom.GetStorableByID("HeadAudioSource").gameObject.AddComponent<AudioHighPassFilter>();
				AFXHPFilter = containingAtom.GetStorableByID("HeadAudioSource").gameObject.GetComponent<AudioHighPassFilter>();
			}			
		}
		
		void InitAFXValues() {
			// First reverb effect
			AFXReverbFilter.enabled = RevEnabled.val;
			AFXReverbFilter.decayTime = RevDecayTime.val;
			AFXReverbFilter.decayHFRatio = RevDecayHFRatio.val;
			AFXReverbFilter.density = RevDensity.val;
			AFXReverbFilter.diffusion = RevDiffusion.val;
			AFXReverbFilter.dryLevel = RevDryLevel.val;
			
			// Then echo effect
			AFXEchoFilter.enabled = EchEnabled.val;
			AFXEchoFilter.decayRatio = EchDecayRatio.val;
			AFXEchoFilter.delay = EchDelay.val;
			AFXEchoFilter.dryMix = EchDryMix.val;
			AFXEchoFilter.wetMix = EchWetMix.val;
			
			// The chorus effect
			AFXChorusFilter.enabled = ChoEnabled.val;
			AFXChorusFilter.delay = ChoDelay.val;
			AFXChorusFilter.depth = ChoDepth.val;
			AFXChorusFilter.dryMix = ChoDryMix.val;
			AFXChorusFilter.rate = ChoRate.val;
			AFXChorusFilter.wetMix1 = ChoWetMix1.val;
			AFXChorusFilter.wetMix2 = ChoWetMix2.val;
			AFXChorusFilter.wetMix3 = ChoWetMix3.val;
			
			// Low pass and High pass
			AFXLPFilter.enabled = LPEnabled.val;
			AFXLPFilter.cutoffFrequency = LPCutoffFreq.val;
			AFXLPFilter.lowpassResonanceQ = LPResonance.val;
			
			AFXHPFilter.enabled = HPEnabled.val;
			AFXHPFilter.cutoffFrequency = HPCutoffFreq.val;
			AFXHPFilter.highpassResonanceQ = HPResonance.val;
		}
		
		// **************************
		// Local Tools
		// **************************
		private void logDebug( string debugText ) {
			SuperController.LogMessage( debugText );
		}
				
		void OnDestroy() {
			try
			{

			}
            catch(Exception e)
            {
                SuperController.LogError("VAMMoanAFX - Exception caught: " + e);
            }
		}
		
		void OnEnable() {
			try
			{
				if( containingAtom != null ) {
					if( containingAtom.GetStorableByID("HeadAudioSource").gameObject.GetComponent<AudioReverbFilter>() == null ) {
						if( containingAtom.type == "Person" ) {
							InitCurrentAudioSource();
						}
						
						InitAFXValues();
					}
				}
			}
            catch(Exception e)
            {
                SuperController.LogError("VAMMoanAFX - Exception caught: " + e);
            }
		}

		// **************************
		// Time to cleanup !
		// **************************		
		
		void OnDisable() {
			try
			{
				// Every single effect on the audio listener / audio source
				if( containingAtom != null ) {
					Destroy( containingAtom.GetStorableByID("HeadAudioSource").gameObject.GetComponent<AudioEchoFilter>() );
					Destroy( containingAtom.GetStorableByID("HeadAudioSource").gameObject.GetComponent<AudioReverbFilter>() );
					Destroy( containingAtom.GetStorableByID("HeadAudioSource").gameObject.GetComponent<AudioChorusFilter>() );
					Destroy( containingAtom.GetStorableByID("HeadAudioSource").gameObject.GetComponent<AudioLowPassFilter>() );
					Destroy( containingAtom.GetStorableByID("HeadAudioSource").gameObject.GetComponent<AudioHighPassFilter>() );
				} else {
				}
			}
            catch(Exception e)
            {
                SuperController.LogError("VAMMoanAFX - Exception caught: " + e);
            }
		}
	}
}